/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id: MessageExchangeSerializer.java 154 2006-03-27 15:30:10Z alouis $
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.tools.rmi.common.serializable;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Properties;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.zip.GZIPInputStream;
import java.util.zip.GZIPOutputStream;

import javax.activation.DataHandler;
import javax.mail.util.ByteArrayDataSource;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.TransformerFactoryConfigurationError;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.stream.StreamSource;



/**
 *
 * @author alouis - EBM Websourcing
 *
 */
public final class MessageExchangeSerializer implements Serializable {

    /**
     * the serialVersionUID.
     */
    private static final long serialVersionUID = 1L;

    /**
     * the pool size.
     */
    private static final int POOLSIZE = 8;

    /**
     * the buffer size.
     */
    private static final int BUFFER_SIZE = 1024;

    /**
     * the singleton.
     */
    private static MessageExchangeSerializer instance =
        new MessageExchangeSerializer();

    /**
     * the transformer pool.
     */
    private ArrayBlockingQueue<Transformer> transformerPool;

    /**
     * the dafult constructor.
     *
     */
    private MessageExchangeSerializer() {
        super();
        // create the transformer pool
        this.transformerPool =
            new ArrayBlockingQueue<Transformer>(MessageExchangeSerializer.POOLSIZE);
        for (int i = 0; i < MessageExchangeSerializer.POOLSIZE; i++) {
            this.transformerPool.add(this.createNewTransformer());
        }
    }


    /**
     * Get the instance.
     * @return
     *       the instance
     */
    public static MessageExchangeSerializer instance() {
        return MessageExchangeSerializer.instance;
    }

    /**
     * Serialize the content.
     * @param content
     *          the content
     * @param s
     *          the object output stream
     * @param compress
     *          the mode
     * @throws IOException
     *          impossible to serialize the content
     * @throws InterruptedException
     *          impossible to serialize the content
     * @throws TransformerException
     *          impossible to serialize the content
     */
    public void serializeContent(final Source content,
            final ObjectOutputStream s,
            final boolean compress)
    throws IOException, InterruptedException, TransformerException {

        synchronized (content) {
            /*
             * Compress content if specified
             */
            StreamResult sresult = null;
            GZIPOutputStream gz = null;
            if (compress) {
                gz = new GZIPOutputStream(s);
                sresult = new StreamResult(gz);

            } else {
                sresult = new StreamResult(s);
            }
            Transformer transform = this.transformerPool.take();
            transform.transform(content, sresult);
            if (content instanceof StreamSource) {
                StreamSource ss = (StreamSource) content;

                ss.getInputStream().reset();
            }
            this.transformerPool.offer(transform);

            if (compress) {
                gz.finish();
            }
        }
    }

    /**
     * Deserialize the content.
     * @param s
     *      the object input stream
     * @param compress
     *       the comporession mode
     * @return
     *        the source
     * @throws IOException
     *          impossible to deserialize the content
     */
    public Source deserializeContent(final ObjectInputStream s,
            final boolean compress)
    throws IOException {
        StreamSource source = new StreamSource();

        InputStream in = s;
        if (compress) {
            in = new GZIPInputStream(s);
        }

        byte[] buff = new byte[MessageExchangeSerializer.BUFFER_SIZE];
        ByteArrayOutputStream outputStream = new ByteArrayOutputStream();
        int n = 0;
        while (-1 != (n = in.read(buff))) {
            outputStream.write(buff, 0, n);
        }

        ByteArrayInputStream result = new ByteArrayInputStream(outputStream
                .toByteArray());

        source.setInputStream(result);
        return source;
    }

    /**
     * Serialize all datahandler in the Map. <br>
     * Serialized information is : <br>
     * -attachmentName <br>
     * -attachmentContentType <br>
     * -name <br>
     * -inputStream available bytes count <br>
     * -inputStream as a byteArray
     *
     * @param attachments
     *          the attachements
     * @param s
     *          the output object stream
     * @throws IOException
     *      impossible to serialze attachments
     */
    public void serializeAttachments(
            final Map<String, DataHandler> attachments,
            final ObjectOutputStream s) throws IOException  {
        s.writeInt(attachments.size());

        for (Iterator iter = attachments.keySet().iterator(); iter.hasNext();) {

            String attachmentName = (String) iter.next();
            DataHandler attachment = attachments.get(attachmentName);

            InputStream binary = attachment.getInputStream();
            synchronized (binary) {

                int available = binary.available();

                // write information about the attachment
                s.writeUTF(attachmentName);
                s.writeUTF(attachment.getContentType());
                s.writeUTF(attachment.getName());
                s.writeInt(available);

                // write attachment content
                byte[] content = new byte[available];
                binary.read(content);
                s.write(content);
                s.flush();
            }
        }
    }

    /**
     * Deserialize all datahandlers in the stream and return a Map. <br>
     * Deserialized information is : <br>
     * -attachmentName <br>
     * -attachmentContentType <br>
     * -name <br>
     * -Stream available bytes count <br>
     * -Stream as a byteArray
     *
     * @param s
     *        the object input stream
     * @return
     *        the deserialized attachments
     * @throws IOException
     *         impossible to deserialize attachments
     */
    public Map<String, DataHandler> deserializeAttachments(
            final ObjectInputStream s)
            throws IOException {
        Map<String, DataHandler> attachments = new HashMap<String, DataHandler>();

        int attachmentsCount = s.readInt();

        for (int i = 0; i < attachmentsCount; i++) {
            // read information
            String attachmentName = s.readUTF();
            String contentType = s.readUTF();
            String name = s.readUTF();
            int available = s.readInt();

            // read content
            byte[] content = new byte[available];
            s.readFully(content);
            ByteArrayDataSource dataSource = new ByteArrayDataSource(content,
                    contentType);
            dataSource.setName(name);

            // create the dataHandler
            DataHandler dataHandler = new DataHandler(dataSource);

            attachments.put(attachmentName, dataHandler);
        }
        return attachments;
    }

    /**
     * Create a new transformer.
     * @return
     *        the transformer
     */
    private Transformer createNewTransformer() {
        Transformer transformer = null;
        try {
            transformer = TransformerFactory.newInstance().newTransformer();
        } catch (TransformerConfigurationException e) {
            e.printStackTrace();
        } catch (TransformerFactoryConfigurationError e) {
            e.printStackTrace();
        }
        Properties props = new Properties();
        props.put(OutputKeys.OMIT_XML_DECLARATION, "yes");
        if (transformer != null) {
            transformer.setOutputProperties(props);
        }
        return transformer;
    }



}
